c
c 
c 
c  Frequency function fit.
c  Owen Long  3/07/98.
c  UCSB
c 
c  Description:
c 
c    Fits data to the functional form:
c 
c     freq( (x-a)/b )  =
c        1 - 1/sqrt(2 pi)  Integral (-inf, to (x-a)/b )  exp( -1/2 y^2 ) dy
c 
c     freq( (x-a)/b ) =  1/2 { 1 - Erf[ (x-a)/b * (1/sqrt(2)) ] }
c 
c    If the data are from a threshold scan at fixed charge injection
c    for the SVT:
c 
c     Each datum is
c       x = Threshold DAC value.  Low values of x correspond to low threshold values.
c       y = Efficiency of channel at the given threshold DAC value (x).
c      dy = Binomial error on y.
c 
c     And the parameters of freq are
c       a = Threshold at which the channel has a hit efficiency of 50%.
c       b = The width of the gaussian that is integrated in the error
c           function which is the measure of the noise.
c 
c 
c    The fit first makes a crude estimate of a and b using the average and
c    difference of two points on the curve as illustrated below.
c 
c       ^
c     y |
c       |
c    1 -|xxxxxxxxxxxxxxxx-----------------
c       |                xx
c       |                  x
c       |                   x-----------  y = MAX_CUT
c       |                   |x
c       |                   |
c       |                   | x
c       |                   |
c   0.5-|                   |
c       |                   |  x
c       |                   |
c       |                   |   x
c       |                   |
c       |                   |    x------  y = MIN_CUT
c       |                   |    |x
c       |                   |    | xx
c    0 - ------------------------------------------------------>
c                           ^    ^
c                           |    |                          x
c                           |   x = minVal
c                           |
c                          x = maxVal
c 
c    where
c          a estimate:  guessThresh =  (minVal+maxVal)/2
c          b estimate:  guessNoise  =  (minVal-maxVal)/WIDTH_FACTOR
c 
c 
c    The best values for a and b are found by minimizing the chisquared
c    using the Newton-Raphson method as described in "Statistics for
c    Nuclear and Particle Physicists" by Louis Lyons (p 136).
c    This involves successive refinements of the estimated a and b
c    values until the step in a and b between refinements is acceptably
c    small.  This is usually achived after 2 to 4 steps.  The fit
c    is not allowed to make more than maxIter steps (nominally set to 10).
c 
c
c ------ Error Codes:
c
c  If the fitchan fails or doesn't even try to fit the data, the 
c   parameters, their errors, and the chisquare/dof are set to a 
c   negative integer.  The following are the error codes:
c
c   Code
c
c    -1    Encountered a negative determinant, tried a brute-force local
c           minimization and failed to find a better point.
c
c    -2    The number of points used in the fit was zero. (never happens?)
c
c    -3    Non-zero number of hits, but below the minimum required.
c
c    -4    Zero hits.  (dead channel)
c
c 
c
      subroutine fitchan(nChip, nChan, a, b, 
     .                   sigA, sigB, chisqPdof, nIter,
     .                   ptsUsed)
      implicit none
      integer nChan, nChip, nIter, ptsUsed
      real    a, b, sigA, sigB, chisqPdof
      include 'quickfit.inc'

      real    yVal, xVal, errSq, tmpA, tmpB
      integer iter
      real    minValue, maxValue
      real    FVal, dFdA, dFdB, d2Fd2A, d2Fd2B, d2FdAdB
      real    SVal, dSdA, dSdB, d2Sd2A, d2Sd2B, d2SdAdB
      real    detE, deltaA, deltaB
      real    tmpSigA, tmpSigB
      real    freq, dFreqDa, dFreqDb, d2FreqD2a, d2FreqD2b, d2FreqDaDb
      real    totalHits, totalArea
      real    guessA, guessB, step
      integer i, ind, errCode, tryNumber, stat

      real    prevChisq
      real    rChisq

c+++++++++++++

      tmpA = a
      tmpB = b

      
      minValue = -1.
      maxValue = -1.


c Get a rough estimate of the threshold and noise from the average
c  and scaled difference of a pair of points.

c Count total number of (rescaled) hits too.
      totalHits = 0.

      do i = 1, 64
         ind = nChip*(128*64) + nChan*64 + i
         yVal = tscan(ind)/maxVal
         xVal = 64 - i
         if (yVal.gt.minCut.and.minValue.eq.-1.) minValue = xVal+1.0
         if (yVal.gt.maxCut.and.maxValue.eq.-1.) maxValue = xVal
         totalHits = totalHits + tscan(ind)/maxVal
      enddo

      guessA = (minValue+maxValue)/2.0
      guessB = (minValue-maxValue)/widthFactor

      tmpA = guessA
      tmpB = guessB

c Must have at minimum of minArea.
c Changed errCode = -4 for no hits at all  (CC 17 mar 98)
c Put in errCode = -5 for channels that are stuck on (ORL 3/18/98).

      totalArea = totalHits / 64.

      errCode = 0.

      if (totalArea.lt.minArea) errCode = -3.
      if (totalHits .lt. 0.001) errCode = -4. 
      if (totalArea.gt.maxArea) errCode = -5.

      if (errCode.lt.0) then
         a =  63. - errCode
         b =  errCode
         sigA = errCode
         sigB = errCode
         chisqPdof = errCode
         nIter = 0
         ptsUsed = 0
         goto 999
      endif


c This is the first try.
      tryNumber = 1


c Begin the fit iteration.

11    continue

      prevChisq = 1e+10

      do iter = 1, maxIter

         SVal = 0.
         dSdA = 0.
         dSdB = 0.
         d2Sd2A = 0.
         d2Sd2B = 0.
         d2SdAdB = 0.

         ptsUsed = 0


c Loop over the threshold values.

         do i = 1, maxThreshDac


            ind = nChip*(128*64) + nChan*64 + i

c Does this point lie in the valid range? (0 to maxVal)

            if (tscan(ind).ge.0.and.tscan(ind).le.maxVal) then


c This data point.

               yVal = tscan(ind)/maxVal

               if (tscan(ind).eq.0.or.tscan(ind).eq.maxVal) then

                  if (useAllPoints) then
c Points with zero binomial error are treated as if they
c  are bins with one (or max-1) hit(s).
                     errSq = (1.0-1.0/maxVal)/(maxVal**2)
                  else
                     errSq = 0.0
                  endif

               else

                  errSq = yVal*(1.0-yVal)/maxVal

               endif
               xVal = 64 - i


c Non-zero error?

               if (errSq.gt.0) then

                  ptsUsed = ptsUsed + 1

c Function and derivative values.


                  FVal = freq(xVal, tmpA, tmpB)
                  dFdA =    dFreqDa(xVal, tmpA, tmpB)
                  dFdB =    dFreqDb(xVal, tmpA, tmpB)
                  d2Fd2A =  d2FreqD2a(xVal, tmpA, tmpB)
                  d2Fd2B =  d2FreqD2b(xVal, tmpA, tmpB)
                  d2FdAdB = d2FreqDaDb(xVal, tmpA, tmpB)

c Chisquare and derivative values.

                  SVal = SVal + ( (yVal - FVal)**2 ) / errSq
                  dSdA = dSdA + 2.0 * ( FVal * dFdA - yVal * dFdA ) /
     .                   errSq
                  dSdB = dSdB + 2.0 * ( FVal * dFdB - yVal * dFdB ) /
     .                   errSq
                  d2Sd2A = d2Sd2A + 2.0 * ( dFdA**2 + FVal * d2Fd2A - 
     .                                   yVal * d2Fd2A ) / errSq
                  d2Sd2B = d2Sd2B + 2.0 * ( dFdB**2 + FVal * d2Fd2B - 
     .                                   yVal * d2Fd2B ) / errSq
                  d2SdAdB = d2SdAdB + 2.0 * ( dFdA * dFdB + FVal * 
     .                      d2FdAdB -     yVal * d2FdAdB ) / errSq

               endif ! Non-zero error.
            endif ! valid range (0 to maxVal).
         enddo ! loop over bins


c Determinant of matrix of second derivatives.

         detE = d2Sd2A * d2Sd2B - d2SdAdB * d2SdAdB
 
         if ( detE.gt.0 .and.
     .        d2Sd2A.gt.0 .and.
     .        d2Sd2B.gt.0        ) then

c Shift from current guess to better values.

            deltaA = (1.0/detE) * ( d2Sd2B * dSdA - d2SdAdB * dSdB )
            deltaB = (1.0/detE) * ( d2Sd2A * dSdB - d2SdAdB * dSdA )

            tmpSigA = sqrt( 2.0 * (1.0/detE) * d2Sd2B )
            tmpSigB = sqrt( 2.0 * (1.0/detE) * d2Sd2A )
 
         else

c Something's wrong.

c Do we want to try again?

            if (tryNumber.le.1) then

c Try inbetween bins that cross 50% point for threshold guess.
               do i = 1, 64
                  ind = nChip*(128*64) + nChan*64 + i
                  yVal = tscan(ind)/maxVal
                  if (yVal.gt.0.50) then
                     guessA = 64 - i - 0.5
                     guessB = 1.0
                     goto 12
                  endif
               enddo
12             continue

               

c Do a rough, brute force local chisquare minimization and try again.

               tmpA = guessA
               tmpB = guessB

               step = 1.0
               call bruteMin( nChip, nChan, tmpA, tmpB, step, 5, stat )
               
               
c Is there a better point?
               if (stat.ne.0) then
c No.
                  errCode = -1
                  a = 63. - errCode
                  b = errCode
                  sigA = errCode
                  sigB = errCode
                  chisqPdof = errCode
                  nIter = 0
                  ptsUsed = 0
               else
                  guessA = tmpA
                  guessB = tmpB
                  tryNumber = tryNumber + 1
                  goto 11
               endif

            endif ! trying again.
 
         endif ! non-zero determinant.


c Break out of loop if shift is small enough.
   
         if (abs(deltaA).lt.tolA.and.
     .       abs(deltaB).lt.tolB) goto 800

c Quit loop if chisquare got worse in this iteration.

         if (SVal.gt.prevChisq) then
            goto 800
         else
            prevChisq = SVal
         endif

c Don't let the noise go negative.

C        if ((tmpB-deltaB).lt.0) then
C           goto 800
C        endif

c Do another iteration.

         tmpA = tmpA - deltaA
         tmpB = tmpB - deltaB

 
      enddo ! loop over iterations

800   continue




c How's the chisquare?
C     if (SVal.gt.chisqCut.and.tryNumber.le.2) then
c Cheat and include the points with zero binomial error (last resort).
C        do i = 1, 64
C           ind = nChip*(128*64) + nChan*64 + i
C           if (tscan(ind).ge.maxVal) tscan(ind) = maxVal - 0.001
C           if (tscan(ind).le.0.0) tscan(ind) =  0.001
C        enddo
C        tmpA = guessA
C        tmpB = guessB
C        tryNumber = tryNumber + 1
C        goto 11
C     endif
         



      if (ptsUsed.gt.0) then

         a = tmpA

c Take the absolute value of tmpB because when it's a really
c  sharp turn-on, the fitter sometimes gives a negative
c  width.
         b = abs(tmpB)

         sigA = tmpSigA
         sigB = tmpSigB

c Calculate the chisquare and number of points used only 
c  using points with non-zero binomial error and threshold
c  below maxThreshDac.
         call realChisq(nChip, nChan, a, b, rChisq, ptsUsed)
         if (ptsUsed.gt.0) then
            chisqPdof = rChisq/ptsUsed
         else
            chisqPdof = 0.
         endif

         nIter = iter
      else
         errCode = -2
         a = 63. - errCode
         b = errCode
         sigA = errCode
         sigB = errCode
         chisqPdof = errCode
         nIter = 0
         ptsUsed = 0
      endif

999   continue
      return
      end

